/**
 * Copyright 2012 OW2 Shelbie
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.ow2.shelbie.core.internal.extension.parser;

import org.apache.felix.ipojo.ConfigurationException;
import org.apache.felix.ipojo.metadata.Attribute;
import org.apache.felix.ipojo.metadata.Element;
import org.ow2.shelbie.core.internal.extension.type.ECommand;
import org.ow2.shelbie.core.internal.extension.type.ECompleter;
import org.ow2.shelbie.core.internal.handler.CommandHandler;
import org.ow2.shelbie.core.internal.handler.CompleterServiceComparator;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.util.Dictionary;
import java.util.List;
import java.util.Map;
import java.util.Vector;

import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertTrue;

/**
 * A {@code CommandParserTestCase} is ...
 *
 * @author Guillaume Sauthier
 */
public class CommandParserTestCase {
    
    private CommandParser parser;
    
    @BeforeMethod
    public void setUp() throws Exception {
        parser = new CommandParser();
    }

    @Test
    public void testCommandWithActionAndNoCompleters() throws Exception {
        Element e = createCommandElement("Action");

        ECommand command = parser.getCommandDescription(e);
        assertEquals(command.getAction(), "Action");
        assertTrue(command.getCompleters().isEmpty());
    }

    @Test
    public void testCommandWithActionAndOneCompleter() throws Exception {
        Element e = createCommandElement("Action");
        createCompleterElement(e, "Completer");

        ECommand command = parser.getCommandDescription(e);
        assertEquals(command.getAction(), "Action");
        assertEquals(command.getCompleters().size(), 1);
        ECompleter completer = command.getCompleters().get(0);
        assertEquals(completer.getComponent(), "Completer");

        // A property is auto-added for the position in the commpleters list
        assertEquals(completer.getConfiguration().get(CompleterServiceComparator.POSITION), "0");
    }

    @Test
    public void testCommandWithActionAndMultipleCompleters() throws Exception {
        Element e = createCommandElement("Action");
        createCompleterElement(e, "Completer");
        createCompleterElement(e, "Completer2");

        ECommand command = parser.getCommandDescription(e);
        
        assertEquals(command.getAction(), "Action");
        assertEquals(command.getCompleters().size(), 2);
        ECompleter completer1 = command.getCompleters().get(0);
        assertEquals(completer1.getComponent(), "Completer");

        // A property is auto-added for the position in the commpleters list
        assertEquals(completer1.getConfiguration().get(CompleterServiceComparator.POSITION), "0");

        ECompleter completer2 = command.getCompleters().get(1);
        assertEquals(completer2.getComponent(), "Completer2");

        // A property is auto-added for the position in the commpleters list
        assertEquals(completer2.getConfiguration().get(CompleterServiceComparator.POSITION), "1");
    }

    @Test
    public void testCommandWithStringProperty() throws Exception {
        Element e = createCommandElement("Action");
        Element c = createCompleterElement(e, "Completer");
        createPropertyElement(c, "hello", "world");

        ECommand command = parser.getCommandDescription(e);
        ECompleter completer = command.getCompleters().get(0);
        assertEquals(completer.getConfiguration().get("hello"), "world");
    }

    @Test
    public void testCommandWithStringArrayProperty() throws Exception {
        Element e = createCommandElement("Action");
        Element c = createCompleterElement(e, "Completer");
        createPropertyElement(c, "hello", "{world,monde}");

        ECommand command = parser.getCommandDescription(e);
        ECompleter completer = command.getCompleters().get(0);
        String[] values = (String[]) completer.getConfiguration().get("hello");
        assertEquals(values[0], "world");
        assertEquals(values[1], "monde");
    }

    @Test
    public void testCommandWithNonEmptyArrayProperty() throws Exception {
        Element e = createCommandElement("Action");
        Element c = createCompleterElement(e, "Completer");
        Element p = createPropertyElement(c, "hello", null);
        p.addAttribute(new Attribute("type", "array"));
        createPropertyElement(p, null, "world");

        ECommand command = parser.getCommandDescription(e);
        ECompleter completer = command.getCompleters().get(0);
        String[] values = (String[]) completer.getConfiguration().get("hello");
        assertEquals(values[0], "world");
    }

    @Test
    public void testCommandWithEmptyArrayProperty() throws Exception {
        Element e = createCommandElement("Action");
        Element c = createCompleterElement(e, "Completer");
        Element p = createPropertyElement(c, "hello", null);
        p.addAttribute(new Attribute("type", "array"));

        ECommand command = parser.getCommandDescription(e);
        ECompleter completer = command.getCompleters().get(0);
        String[] values = (String[]) completer.getConfiguration().get("hello");
        assertEquals(values.length, 0);
    }

    @Test
    public void testCommandWithNonEmptyVectorProperty() throws Exception {
        Element e = createCommandElement("Action");
        Element c = createCompleterElement(e, "Completer");
        Element p = createPropertyElement(c, "hello", null);
        p.addAttribute(new Attribute("type", "vector"));
        createPropertyElement(p, null, "world");

        ECommand command = parser.getCommandDescription(e);
        ECompleter completer = command.getCompleters().get(0);
        Vector<String> values = (Vector<String>) completer.getConfiguration().get("hello");
        assertEquals(values.get(0), "world");
    }

    @Test
    public void testCommandWithEmptyVectorProperty() throws Exception {
        Element e = createCommandElement("Action");
        Element c = createCompleterElement(e, "Completer");
        Element p = createPropertyElement(c, "hello", null);
        p.addAttribute(new Attribute("type", "vector"));

        ECommand command = parser.getCommandDescription(e);
        ECompleter completer = command.getCompleters().get(0);
        Vector<String> values = (Vector<String>) completer.getConfiguration().get("hello");
        assertTrue(values.isEmpty());
    }

    @Test
    public void testCommandWithNonEmptyListProperty() throws Exception {
        Element e = createCommandElement("Action");
        Element c = createCompleterElement(e, "Completer");
        Element p = createPropertyElement(c, "hello", null);
        p.addAttribute(new Attribute("type", "list"));
        createPropertyElement(p, null, "world");

        ECommand command = parser.getCommandDescription(e);
        ECompleter completer = command.getCompleters().get(0);
        List<String> values = (List<String>) completer.getConfiguration().get("hello");
        assertEquals(values.get(0), "world");
    }

    @Test
    public void testCommandWithEmptyListProperty() throws Exception {
        Element e = createCommandElement("Action");
        Element c = createCompleterElement(e, "Completer");
        Element p = createPropertyElement(c, "hello", null);
        p.addAttribute(new Attribute("type", "list"));

        ECommand command = parser.getCommandDescription(e);
        ECompleter completer = command.getCompleters().get(0);
        List<String> values = (List<String>) completer.getConfiguration().get("hello");
        assertTrue(values.isEmpty());
    }

    @Test
    public void testCommandWithNonEmptyMapProperty() throws Exception {
        Element e = createCommandElement("Action");
        Element c = createCompleterElement(e, "Completer");
        Element p = createPropertyElement(c, "hello", null);
        p.addAttribute(new Attribute("type", "map"));
        createPropertyElement(p, "hello", "world");
        createPropertyElement(p, "bonjour", "monde");

        ECommand command = parser.getCommandDescription(e);
        ECompleter completer = command.getCompleters().get(0);
        Map<String, String> values = (Map<String, String>) completer.getConfiguration().get("hello");
        assertEquals(values.get("hello"), "world");
        assertEquals(values.get("bonjour"), "monde");
    }

    @Test
    public void testCommandWithEmptyMapProperty() throws Exception {
        Element e = createCommandElement("Action");
        Element c = createCompleterElement(e, "Completer");
        Element p = createPropertyElement(c, "hello", null);
        p.addAttribute(new Attribute("type", "map"));

        ECommand command = parser.getCommandDescription(e);
        ECompleter completer = command.getCompleters().get(0);
        Map<String, String> values = (Map<String, String>) completer.getConfiguration().get("hello");
        assertTrue(values.isEmpty());
    }

    @Test
    public void testCommandWithNonEmptyDictionaryProperty() throws Exception {
        Element e = createCommandElement("Action");
        Element c = createCompleterElement(e, "Completer");
        Element p = createPropertyElement(c, "hello", null);
        p.addAttribute(new Attribute("type", "dictionary"));
        createPropertyElement(p, "hello", "world");
        createPropertyElement(p, "bonjour", "monde");

        ECommand command = parser.getCommandDescription(e);
        ECompleter completer = command.getCompleters().get(0);
        Dictionary<String, String> values = (Dictionary<String, String>) completer.getConfiguration().get("hello");
        assertEquals(values.get("hello"), "world");
        assertEquals(values.get("bonjour"), "monde");
    }

    @Test
    public void testCommandWithEmptyDictionaryProperty() throws Exception {
        Element e = createCommandElement("Action");
        Element c = createCompleterElement(e, "Completer");
        Element p = createPropertyElement(c, "hello", null);
        p.addAttribute(new Attribute("type", "dictionary"));

        ECommand command = parser.getCommandDescription(e);
        ECompleter completer = command.getCompleters().get(0);
        Dictionary<String, String> values = (Dictionary<String, String>) completer.getConfiguration().get("hello");
        assertTrue(values.isEmpty());
    }

    @Test(expectedExceptions = ConfigurationException.class)
    public void testCommandWithUnknownPropertyType() throws Exception {
        Element e = createCommandElement("Action");
        Element c = createCompleterElement(e, "Completer");
        Element p = createPropertyElement(c, "hello", null);
        p.addAttribute(new Attribute("type", "unknown"));

        parser.getCommandDescription(e);
    }

    private Element createCommandElement(String action) {
        Element e = new Element(CommandHandler.COMMAND_ELEMENT, CommandHandler.NAMESPACE);
        e.addAttribute(new Attribute("action", action));
        return e;
    }

    private Element createCompleterElement(Element command, String componentName) {
        Element c = new Element("completer", CommandHandler.NAMESPACE);
        c.addAttribute(new Attribute("component", componentName));
        command.addElement(c);
        return c;
    }

    private Element createPropertyElement(Element completer, String name, String value) {
        Element p = new Element("property", null);
        if (name != null) {
            p.addAttribute(new Attribute("name", name));
        }
        if (value != null) {
            p.addAttribute(new Attribute("value", value));
        }
        completer.addElement(p);
        return p;
    }

}
